------------------------------
--      Are you local?      --
------------------------------

local L = AceLibrary("AceLocale-2.2"):new("BagginsAnywhereBags")

local _G = _G
local type = _G.type
local pairs = _G.pairs
local tonumber = _G.tonumber
local select = _G.select
local playerName = UnitName("player")

local r,p

----------------------------
--      Localization      --
----------------------------

L:RegisterTranslations('enUS', function() return {	
	["Anywhere Bags"] = true,
	["View any character's bank and bags from anywhere."] = true,
	["Compressed"] = true,
	["Search"] = true,
	["Purge"] = true,
	["<itemLink, itemName, itemID>"] = true,
	["|cff7fff7fSearching all players for:|r "] = true,
	["|cffff0000No items found.|r"] = true,
	["Bags"] = true,
	["Equiped"] = true,
	["Keyring"] = true,
	["Mail"] = true,
	["Unknown id"] = true,
} end)

---------------------------------
--      Addon Declaration      --
---------------------------------
--BAB = BagginsAnywhereBags
BagginsAnywhereBags = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceDebug-2.0", "AceHook-2.1")
local BagginsAnywhereBags = BagginsAnywhereBags
BagginsAnywhereBags.revision = tonumber(string.sub("$Revision: 47373 $", 12, -3))

---------------------------------
--      AceOptions Config      --
---------------------------------

local options = {
	name = L["Anywhere Bags"], type='group',
	desc = L["View any character's bank and bags from anywhere."],
	handler = BagginsAnywhereBags,
	order = 200,
	wfHidden = true,
	cmdHidden = true,
	args = {
		search = {
			name = L["Search"], type='text',
			desc = L["Search"],
			get = false,
			set = "Search",
			usage = L["<itemLink, itemName, itemID>"],
			validate = function(v) return true end,
			order = -40,
		},
		purge = {
			name = L["Purge"], type='group',
			desc = L["Purge"],
			pass = true,
			func = "PurgePlayer",
			args = {},
			order = -35,
		},
		spacer = {
			type='header',
			order = -30,
		},
		compressed = {
			name = L["Compressed"], type='toggle',
			desc = L["Compressed"],   
			get = function() return BagginsAnywhereBags.db.profile.compressed end,
			set = "SetCompressed",
			order = -10,
		},
	},
}

------------------------------
--      Initialization      --
------------------------------

function BagginsAnywhereBags:OnInitialize()
	self:RegisterDB("BagginsAnywhereBagsDB")
	self:RegisterDefaults('realm', {
		[playerName] = {
			[-2] = {}, [-1] = {}, [0] = {};
			-- 1 to 11:
			{},{},{},{},{},{},{},{},{},{},{};
			Equiped = {},
			Money = 0,
			Mail = {},
		},
		DBVersion = self.revision, --TODO: testing code, remove later
	})
	self:RegisterDefaults('profile', {
		compressed = true,
		DBVersion = self.revision, --TODO: testing code, remove later
	})
	self:RegisterChatCommand({"/BagginsAnywhereBags", "/BAB"}, {
		name = options.name, type='group',
		desc = options.desc,
		args = {
			search = options.args.search,
			purge = options.args.purge,
			compressed = options.args.compressed,
		},
	})
end

--[[----------------
items stored as:
BagginsAnywhereBags.db.realm[player][bag][slot] = {itemLink, count, texture, quality, Baggins_bagid, Baggins_sectionid}

itemLink regex:
itemstring, itemname = |H(.+)|h, |h(.+)|h
----------------]]--

function BagginsAnywhereBags:OnEnable(first)
	if not Baggins then return end
	
	r = self.db.realm
	p = self.db.profile
	
	if first then
		--fix databases here
		if type(p.DBVersion) ~= 'number' then
			p.DBVersion = 3
		end
		if p.DBVersion < 4 then
			p.sortType = nil
			p.useFrames = nil
			p.DBVersion = 4
		end
		if p.DBVersion < 47189 then
			p.DBVersion = 47189
		end
		--
		if type(r.DBVersion) ~= 'number' then
			r.DBVersion = 3
		end
		if r.DBVersion < 4 then
			for player,bags in pairs(r) do
				if type(bags) == 'table' then
					for bag,slots in pairs(bags) do
						if bag == "Mail" or bag == -2 or bag == "Equiped" then
							for slot,item in pairs(slots) do
								item[5] = bag == -2 and L["Keyring"] or bag == "Mail" and L["Mail"] or bag == "Equiped" and L["Equiped"] or itemp[5]
								item[6] = nil
							end
						end
					end
				end
			end
			r.DBVersion = 4
		end
		if r.DBVersion < 47189 then
			r.DBVersion = 47189
		end
		--
	end
	
	--bank and mail need stored whenever they change
	self:RegisterEvent("BANKFRAME_CLOSED") -- fired twice
	self:RegisterEvent("MAIL_CLOSED") -- fired twice
	
	--only care about storing bags when logging out (you can always access your own bags)
	if type(_G["Logout"]) == 'function' then
		self:Hook("Logout", "OnLogout", true)
	end
	
	if type(_G["Quit"]) == 'function' then
		self:Hook("Quit", "OnQuit", true)
	end
	
	self:RebuildOptions()
	
	Baggins.OnMenuRequest.args.anywherebags = options
end

function BagginsAnywhereBags:OnDisable()
	if not Baggins then return end
	Baggins.OnMenuRequest.args.anywherebags = nil
end

------------------------------
--      Core Functions      --
------------------------------

local function BagDisabled(player, bag)
	return player == playerName and bag ~= L["Mail"] and not (tonumber(bag) and (bag == -1 or bag > 4 and bag < 12))
end

function BagginsAnywhereBags:ProcessBag(bag)
	r[playerName][bag] = {}

	for i = 1, GetContainerNumSlots(bag) do
		local texture, count = GetContainerItemInfo(bag, i);
		if texture then
			local link = GetContainerItemLink(bag, i)
			r[playerName][bag][i] = {
				link,
				count,
				texture,
				select(3,GetItemInfo(link)),
				self:GetBagginsIDs(bag, i)
			}
		else
			r[playerName][bag][i] = nil
		end
	end
end

function BagginsAnywhereBags:ProcessEquiped()
	for i = 0, 19 do
		local texture = GetInventoryItemTexture("player",i)
		if texture then
			r[playerName].Equiped[i] = {
				GetInventoryItemLink("player", i),
				GetInventoryItemCount("player",i),
				texture,
				GetInventoryItemQuality("player",i),
				L["Equiped"],
			}
		else
			r[playerName].Equiped[i] = nil
		end
	end
end

function BagginsAnywhereBags:OnQuit(...)
	if not pcall(self.FullProcess,self) then
		self:Debug("Error in FullProcess.")
	end
	self.hooks["Quit"](...)
end

function BagginsAnywhereBags:OnLogout(...)
	if not pcall(self.FullProcess,self) then
		self:Debug("Error in FullProcess.")
	end
	self.hooks["Logout"](...)
end

function BagginsAnywhereBags:FullProcess()
	if Baggins.db.profile.highlightnew and Baggins.GetCategoryCache 
			and Baggins:GetCategoryCache().New and next(Baggins:GetCategoryCache().New) then
		Baggins:SaveItemCounts()
		Baggins:ForceFullUpdate()
	end
	self:ProcessBag(-2)
	for i = 0,4 do
		self:ProcessBag(i)
	end
	self:ProcessEquiped()
	r[playerName].Money = GetMoney()
end

function BagginsAnywhereBags:GetBagginsIDs(bag, slot)
	if bag == -2 then
		return L["Keyring"]
	end
	
	local key = bag..":"..slot
	
	local IsSlotMine = function(mybagid, mysecid, key)
		for bagid, bagframe in ipairs(Baggins.bagframes) do
			for sectionid, section in ipairs(bagframe.sections) do
				if section.slots[key] then
					return mybagid == bagid and mysecid == sectionid
				end
			end
		end
	end
	
	for bagid, bag in pairs(Baggins.db.profile.bags) do
		for sectionid, section in pairs(bag.sections) do
			if IsSlotMine(bagid, sectionid, key) then
				return bag.name, section.name
			end
		end
	end
	
	return L["Unknown id"]
end

function BagginsAnywhereBags:Search(v)
	local found = false
	local counts = {}
	
	DEFAULT_CHAT_FRAME:AddMessage(L["|cff7fff7fSearching all players for:|r "]..v)
	
	for player, bags in pairs(r) do
		if type(bags) == 'table' then
			for bag,slots in pairs(bags) do
				if type(slots) == 'table' then
					for slot,item in pairs(slots) do
						if type(item) == 'table' and not BagDisabled(player, bag) then
							if string.lower(item[1]):find(string.lower(v:match("|h(.+)|h") or v),1,true) then
								local itemName = item[1]:match("|h(.+)|h")
								if type(counts[player]) ~= 'table' then
									counts[player] = {}
								end
								if type(counts[player][itemName]) ~= 'table' then
									counts[player][itemName] = {"|cffffff00["..player.."]:|r "..item[1].."x", item[2]}
								else
									counts[player][itemName][2] = counts[player][itemName][2] + item[2]
								end
								found = true
							end
						end
					end
				end
			end
		end
	end
	if found then	
		for _,v1 in pairs(counts) do
			for _,v in pairs(v1) do
				DEFAULT_CHAT_FRAME:AddMessage(v[1]..v[2])
			end
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage(L["|cffff0000No items found.|r"])
	end
end

function BagginsAnywhereBags:PurgePlayer(v)
	r[v] = nil
	options.args.purge.args[v] = nil
	options.args[v] = nil
end

function BagginsAnywhereBags:SetCompressed(v)
	p.compressed = v 
	self:RebuildOptions()
end

-------------------------------
--      Event Functions      --
-------------------------------

do --function BagginsAnywhereBags:BANKFRAME_CLOSED()
	local bankAccess = true --only want to access every other one
	function BagginsAnywhereBags:BANKFRAME_CLOSED()
		bankAccess = not bankAccess
		if bankAccess then return end
		self:ProcessBag(-1)
		for i = 5,11 do
			self:ProcessBag(i)
		end
		self:RebuildBankOptions()
	end
end

do --function BagginsAnywhereBags:MAIL_CLOSED()
   local mailAccess = true --only want to access every other one
   function BagginsAnywhereBags:MAIL_CLOSED()
      mailAccess = not mailAccess
      if mailAccess then return end
      
      r[playerName][L["Mail"]] = {}
      
      for i = 1,GetInboxNumItems() do
         local link = GetInboxItemLink(i)
         if link then
            local count = select(3,GetInboxItem(i))
            local quality,_,_,_,_,_,_,texture = select(3,GetItemInfo(link))
            tinsert(r[playerName][L["Mail"]], {
					link,
					count,
					texture,
					quality,
					L["Mail"],
				})
         end
      end
		if #r[playerName][L["Mail"]] > 0 then
			self:RebuildMailOptions()
		end
   end
end

---------------------------------
--      Display Functions      --
---------------------------------

function FormatMoney(money)
  if money <= 0 then return end
  
  local g,s,c = money/10000, money%10000/100, money%100
  
  local str = ""
  
  if money > 9999 then
    str = str..format("%d|cffffd700g |r",g)
  end
  if money > 99 then
    str = str..format("%d|cffc7c7cfs |r",s)
  end
  if money > 0 then
    str = str..format("%d|cffeda55fc |r",c)
  end
  
  return str
end

function BagginsAnywhereBags:AddBag(player,bag,slots)
	for slot,item in pairs(slots) do
		local itemLink, count, bagid, sectionid = item[1], item[2], item[5], item[6]
		if type(options.args[player].args[bagid]) ~= 'table' then
			options.args[player].args[bagid] = {
				name = bagid, type='group',
				desc = bagid,
				args = {},
				disabled = BagDisabled(player, bag) and not self:IsDebugging()
			}
		end
		local opBag = options.args[player].args[bagid].args
		if sectionid then
			if type(opBag[sectionid]) ~= 'table' then
				opBag[sectionid] = {
					name = sectionid, type='group',
					desc = sectionid,
					args = {},
				}
			end
			opBag = opBag[sectionid].args
		end
		local itemName = p.compressed and itemLink:match("|h(.+)|h") or itemLink
		if type(opBag[itemName]) == 'table' and p.compressed then
			opBag[itemName].count = opBag[itemName].count + count
			opBag[itemName].name = itemLink.."x"..(opBag[itemName].count)
			if self:IsDebugging() then
				opBag[itemName].desc = opBag[itemName].desc..("\n{%s, %s; %s, %s}"):format(bag,slot,bagid,sectionid or "")
			else
				opBag[itemName].desc = opBag[itemName].name
			end
		else
			opBag[itemName] = {
				name = itemLink.."x"..count, type='execute',
				desc = itemLink.."x"..count,
				func = function() SetItemRef(itemLink, itemLink, "LeftButton") end, -- thanks Elkano!
				icon = item[3],
				order = 106-item[4],
				count = count,
			}
			if self:IsDebugging() then
				opBag[itemName].desc = itemLink..("\n{%s, %s; %s, %s}"):format(bag,slot,bagid,sectionid or "")
			end
		end
	end
end

local dummyFunc = function() end
function BagginsAnywhereBags:RebuildOptions()
	for player,bags in pairs(r) do
		if type(bags) == 'table' then --TODO: testing code, remove check later (Checking because of DBVersion)
			--purge
			options.args.purge.args[player] = {
				name = player, type='execute',
				desc = player,
			}
			--
			options.args[player] = {
				name = player, type='group',
				desc = player,
				args = {},
			}
			for bag,slots in pairs(bags) do
				if type(slots) == 'table' then
					self:AddBag(player, bag, slots)
				elseif player ~= playerName then
					local m = FormatMoney(slots)
					options.args[player].args.Money = {
						name = m, type='execute',
						desc = m,
						func = dummyFunc,
						order = -1,
					}
				end
			end
		end --TODO: testing code, remove later
	end
end

function BagginsAnywhereBags:RebuildBankOptions()
	for bagid,bag in pairs(options.args[playerName].args) do
		if not bag.disabled and bagid ~= "Money" and bagid ~= "Mail" then
			options.args[playerName].args[bagid] = nil
		end
	end
	for bag,slots in pairs(r[playerName]) do
		if tonumber(bag) and (bag == -1 or bag > 4 and bag < 12) then
			self:AddBag(playerName, bag, slots)
		end
	end
end

function BagginsAnywhereBags:RebuildMailOptions()
	options.args[playerName].args[L["Mail"]] = nil
	self:AddBag(playerName, "Mail", r[playerName][L["Mail"]])
end
